// -*- C++ -*-
/*!
 * @file  rtc_3DMapVRMLConverter.h
 * @brief Process of converting an object file to VRML
 * @date  $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#ifndef RTC_3DMAPVRMLCONVERTER_H
#define RTC_3DMAPVRMLCONVERTER_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

//Add to code
#include "local_basic_type.h"
#include "opencv2/opencv.hpp"
#include <list>
#include <Shlwapi.h>
#include <direct.h>
#include <time.h>
#include <iomanip>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics.hpp>

using namespace boost::accumulators;

//Add to code
#define _ON_INITIALIZE                 ("onInitialize execution")
#define _ON_ACTIVATED                  ("onActivated execution")
#define _ON_DEACTIVATED                ("onDeactivated execution")
#define _ON_ABORTING                   ("onAborting execution")
#define _ON_RESET                      ("onReset execution")

#define _ERROR_RESET                   ("ERROR: Error was detected. Please reset the 3DMapFindCorrespondence RTC by the RT System Editor.")
#define _STAGE_ERROR                   ("ERROR: Wrong value of the stage variable")
#define _ERROR_FILE_NAME               ("ERROR: The input file path is not found")
#define _ERROR_FILE_CNT                ("ERROR: The number of input files is incorrect")
#define _ERROR_FILE_EXT                ("ERROR: This input extension is invalid")
#define _ERROR_FILE_FORMAT_CALC        ("ERROR: Calculation result file format is invalid")
#define _ERROR_FILE_READ_CALC          ("ERROR: Failed to read the calculation result file")
#define _ERROR_FILE_FORMAT_OBJ         ("ERROR: Object data format is invalid")
#define _ERROR_FILE_READ_OBJ           ("ERROR: Failed to read the 3D object file")
#define _ERROR_FILE_MISMAT_OBJ         ("ERROR: Mismatch of the number of 3D data")
#define _ERROR_AXIS_RANGE              ("ERROR: This value of axis is out of range")
#define _ERROR_FILE_RELEVANCE          ("ERROR: The relevance of file  is incorrect")
#define _ERROR_CREATE_DIRECTORY        ("ERROR: Can't create output file directory")
#define _ERROR_FILE_FORMAT_IMG         ("ERROR: Image data format is invalid")
#define _ERROR_FILE_NOT_COLOR          ("ERROR: Only a color image is effective")
#define _ERROR_FILE_SAVE_IMG           ("ERROR: Can't create texture file")
#define _ERROR_FILE_SAVE_VRML          ("ERROR: Can't create output file of 3DMapVRMLConverter")

#define _VRMLCONVERT_SUCCESS           ("3DMapVRMLConverter RTC process was successful")
#define _VRMLCONVERT_FAILED            ("3DMapVRMLConverter RTC process was failed")

#define _VRMLCONV_STAGE_INIT           (0)
#define _VRMLCONV_STAGE_MIN            (_VRMLCONV_STAGE_INIT)
#define _VRMLCONV_STAGE1               (1)
#define _VRMLCONV_STAGE2               (2)
#define _VRMLCONV_STAGE3               (3)
#define ERROR_DETECTED                  _VRMLCONV_STAGE3
#define _VRMLCONV_STAGE_MAX            (_VRMLCONV_STAGE3+1)

#define _OUTPUT_DIRECTORY              ("C:\\VRML")

#define FILE_PATH_MODE_DIR             (0)
#define FILE_PATH_MODE_EXT             (1)
#define FILE_PATH_MODE_FIL             (2)
#define FILE_PATH_MODE_FNM             (3)

#define DEF_MAX_LENGTH                 (1024)
#define DEF_MID_LENGTH                 (512)
#define DEF_MIN_LENGTH                 (64)
#define DEF_GET_PATH_NG                (0)
#define DEF_GET_PATH_OK                (1)

#define _CHKP_CALC_TRG                 ("(CALC result file path)")
#define _CHKP_CALC_MIN                 (1)
#define _CHKP_CALC_MAX                 (9)
#define _CHKP_CALC_EXT                 (".calcresult")
#define _CHKP_OBJ_TRG                  ("(Object file path)")
#define _CHKP_OBJ_MIN                  (2)
#define _CHKP_OBJ_MAX                  (10)
#define _CHKP_OBJ_EXT                  (".obj")
#define _CHKP_IMG_TRG                  ("(Image file path)")
#define _CHKP_IMG_MIN                  (2)
#define _CHKP_IMG_MAX                  (10)
#define _CHKP_IMG_EXT1                 (".bmp")
#define _CHKP_IMG_EXT2                 (".dib")
#define _CHKP_IMG_EXT3                 (".jpe")
#define _CHKP_IMG_EXT4                 (".jpeg")
#define _CHKP_IMG_EXT5                 (".jpg")
#define _CHKP_IMG_EXT6                 (".png")
#define _CHKP_IMG_EXT7                 (".tif")
#define _CHKP_IMG_EXT8                 (".tiff")

#define _DELEMITER                     (' ')
#define _COMMENT_MARK                  ("#")
#define _FIELDS_CNT_CALC               (7)
#define _FIELDS_CNT_OBJ_VT             (3)
#define _FIELDS_CNT_OBJ_V              (4)
#define _LABEL_VT                      ("vt")
#define _LABEL_V                       ("v")

#define XIMG_120                       (120)
#define YIMG_90                        (90)
#define XIMG_AF_CONV                   (120)
#define YIMG_AF_CONV                   (90)
#define COMPRESS_RATIO                 (static_cast<float_type>XIMG_AF_CONV/XIMG_120)

#define CHANNEL_NUM                    (3)
#define LOW_LUMINANCE                  (0)
#define HIGH_LUMINANCE                 (255)
#define RGB_B                          (0)
#define RGB_G                          (1)
#define RGB_R                          (2)

#define _EMPTY_STRING                  ("")
#define _FLOAT_ZERO                    (0.0)
#define _FLOAT_ONE                     (1.0)

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

/*!
 * @class rtc_3DMapVRMLConverter
 * @brief Process of converting an object file to VRML
 *
 * Image processing of the 3DMapVRMLConverter
 *
 * file path information
 *
 */
class rtc_3DMapVRMLConverter
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_3DMapVRMLConverter(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_3DMapVRMLConverter();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of the 3DMapVRMLConverter RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * 
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the 3DMapVRMLConverter RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the 3DMapVRMLConverter RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the 3DMapVRMLConverter RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the 3DMapVRMLConverter RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the 3DMapVRMLConverter RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::TimedStringSeq m_in_calcfilepath;
  /*!
   * This port inputs the file path of the calculation data.
   * - Type: RTC::TimedStringSeq
   * - Number: 1
   * - Frequency: Caused by user actions.
   * - Operation Cycle: 1 cycles per second.
   */
  InPort<RTC::TimedStringSeq> m_in_calcfilepathIn;
  RTC::TimedStringSeq m_in_objfilepath;
  /*!
   * This port inputs the file path of  the 3D object.
   * - Type: RTC::TimedStringSeq
   * - Number: 1
   * - Frequency: Caused by user actions.
   * - Operation Cycle: 1 cycles per second.
   */
  InPort<RTC::TimedStringSeq> m_in_objfilepathIn;
  RTC::TimedStringSeq m_in_imgfilepath;
  /*!
   * This port inputs the file path of the image.
   * - Type: RTC::TimedStringSeq
   * - Number: 1
   * - Frequency: Caused by user actions.
   * - Operation Cycle: 1 cycles per second.
   */
  InPort<RTC::TimedStringSeq> m_in_imgfilepathIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>
     
  // Add to code
  struct CHK_FILE_PARA
  {
    std::string chk_target;
    ubyte_type min_seq_cnt;
    ubyte_type max_seq_cnt;
    std::list<std::string> list_ext;
  };

  struct CALC_RES
  {
    float_type af_x;
    float_type af_y;
    float_type var_z1;
    float_type mean_z1;
    float_type var_z2;
    float_type mean_z2;
    float_type dc;      // depth coefficient
  };

  struct PIC
  {
    ubyte_type x;
    ubyte_type y;
    float_type z;
  };

  struct LBL_VT
  {
    float_type vt_x;
    float_type vt_y;
  };

  struct LBL_V
  {
    float_type v_x;
    float_type v_y;
    float_type v_z;
  };

  struct VRML_Z
  {
    uword_type length_x;
    uword_type length_y;
    float_type z[XIMG_AF_CONV][YIMG_AF_CONV];
  };

  struct DC_ACCUMULATE
  {
    float_type stdev;        
    float_type mean;
    float_type dc;    // depth coefficient
  };

  CALC_RES m_calc_res[_CHKP_CALC_MAX];
  std::vector<LBL_VT> m_vec_vt[_CHKP_OBJ_MAX];
  std::vector<LBL_V> m_vec_v[_CHKP_OBJ_MAX];
  std::string m_img_file_path[_CHKP_IMG_MAX];

  uword_type m_calc_cnt;
  uword_type m_obj_cnt;
  uword_type m_img_cnt;

  bool_type m_calc_file_flag;
  bool_type m_obj_file_flag;
  bool_type m_img_file_flag;

  ubyte_type m_ubStageCtrl;
  ubyte_type m_ubStageCtrlBackUp;

  /* Common */
  void nvdVRMLConvOnInitialize(void);
  void nvdVRMLConvOnActivated(void);
  void nvdVRMLConvOnDeactivated(void);
  void nvdVRMLConvOnAborting(void);
  void nvdVRMLConvOnReset(void);
  void nvdVRMLConvOnExecute(void);
  void nvdClrMultipleFlags(void);
  void nvdSetStageControl(ubyte_type);
  void nvdClrCalcResult(void);
  void nvdClrObjectData(void);
  void nvdClrImageFile(void);
  bool_type bCheckFileExist(std::string);
  bool_type bCheckExtension(std::string, std::list<std::string> *);
  std::string sEditFilePath(std::string, uword_type);
  std::vector<std::string> split(std::string &, char_type);

  /* Stage1 */
  void nvdCheckofCalcFile(void);
  void nvdCheckofObjFile(void);
  void nvdCheckofImgFile(void);
  uword_type uwVRMLConvOnExecuteStage1(void);
  uword_type uwCheckFileRelevance(void);
  bool_type bCheckInputFile(RTC::TimedStringSeq *, CHK_FILE_PARA  *);
  bool_type bSetCalcResult(const RTC::TimedStringSeq *);
  bool_type bReadCalcResultFile(const std::string, CALC_RES *);
  bool_type bCalcResultSplitter(std::ifstream *, CALC_RES *);
  bool_type bSetObjectData(const RTC::TimedStringSeq *);
  bool_type bReadObjectDataFile(const std::string, std::vector<LBL_VT> *, std::vector<LBL_V> *);
  bool_type bObjectDataSplitter(std::ifstream *, std::vector<LBL_VT> *, std::vector<LBL_V> *);
  bool_type bCheckObjectData(const std::vector<LBL_VT> *);
  bool_type bSetImageFile(const RTC::TimedStringSeq *);

  /* Stage2 */
  void nvdFittingObject(const std::vector<LBL_VT>, const std::vector<LBL_V>, VRML_Z *);
  void nvdTrimmingObject(const VRML_Z *, VRML_Z *);
  void nvdSmooth(VRML_Z *);
  void nvdCalcAccumulate(const VRML_Z *, DC_ACCUMULATE *);
  void nvdCheckAndConvertEnantiomorphism(const VRML_Z *, VRML_Z *, const float_type, const float_type, CALC_RES *);
  void nvdReverseDepth(VRML_Z *, float_type);
  void nvdCorrection(const uword_type, VRML_Z *, DC_ACCUMULATE[]);
  uword_type uwVRMLConvOnExecuteStage2(void);
  bool_type bCreateDirectory(const std::string);
  bool_type bEditSaveTexture(const std::string, const std::string);
  bool_type bSaveVRMLFile(const std::string, const float_type, const float_type, const VRML_Z *, const std::string);
  bool_type bSearchBlueback(const cv::Mat *, uword_type *, uword_type *, uword_type *, uword_type *);
  bool_type bIsEnantiomorphismPixel(const float_type, const float_type, const float_type, const float_type);

};


extern "C"
{
  DLL_EXPORT void rtc_3DMapVRMLConverterInit(RTC::Manager* manager);
};

#endif // RTC_3DMAPVRMLCONVERTER_H
