// -*- C++ -*-
/*!
 * @file  rtc_3DMapVRMLConverter.cpp
 * @brief Process of converting an object file to VRML
 * @date $Date$
 *
 * @author East Japan Accounting Center Co.,Ltd.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#include "rtc_3DMapVRMLConverter.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_3dmapvrmlconverter_spec[] =
  {
    "implementation_id", "rtc_3DMapVRMLConverter",
    "type_name",         "rtc_3DMapVRMLConverter",
    "description",       "Process of converting an object file to VRML",
    "version",           "1.0.0",
    "vendor",            "EAC",
    "category",          "Image",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
rtc_3DMapVRMLConverter::rtc_3DMapVRMLConverter(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_calcfilepathIn("CalcFilePath", m_in_calcfilepath),
    m_in_objfilepathIn("ObjFilePath", m_in_objfilepath),
    m_in_imgfilepathIn("ImgFilePath", m_in_imgfilepath)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
rtc_3DMapVRMLConverter::~rtc_3DMapVRMLConverter()
{
}


/*!
 * Initialization of the 3DMapVRMLConverter RTC
 */
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("CalcFilePath", m_in_calcfilepathIn);
  addInPort("ObjFilePath", m_in_objfilepathIn);
  addInPort("ImgFilePath", m_in_imgfilepathIn);
  
  // Set OutPort buffer
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  
  // add to code
  nvdVRMLConvOnInitialize();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Activate the 3DMapVRMLConverter RTC
 */

RTC::ReturnCode_t rtc_3DMapVRMLConverter::onActivated(RTC::UniqueId ec_id)
{
  // add to code
  nvdVRMLConvOnActivated();

  return RTC::RTC_OK;
}

/*!
 * Deactivate the 3DMapVRMLConverter RTC
 */

RTC::ReturnCode_t rtc_3DMapVRMLConverter::onDeactivated(RTC::UniqueId ec_id)
{
  // add to code
  nvdVRMLConvOnDeactivated();

  return RTC::RTC_OK;
}

/*!
 * Periodic task of the 3DMapVRMLConverter RTC
 */

RTC::ReturnCode_t rtc_3DMapVRMLConverter::onExecute(RTC::UniqueId ec_id)
{
  // add to code
  nvdVRMLConvOnExecute();

  return RTC::RTC_OK;
}

/*!
 * onAborting the 3DMapVRMLConverter RTC
 */

RTC::ReturnCode_t rtc_3DMapVRMLConverter::onAborting(RTC::UniqueId ec_id)
{
  // add to code
  nvdVRMLConvOnAborting();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Reset initialization of the 3DMapVRMLConverter RTC
 */

RTC::ReturnCode_t rtc_3DMapVRMLConverter::onReset(RTC::UniqueId ec_id)
{
  // add to code start
  nvdVRMLConvOnReset();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DMapVRMLConverter::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/***************************************************************************************************
*  Function        : nvdVRMLConvOnExecute
*  Description     : This function execute the main process of 3DMapVRMLConverter RTC.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnExecute(void)
{
  uword_type auwReturn = RESULT_NG;

  switch (m_ubStageCtrl)
  {
  case _VRMLCONV_STAGE1:
    auwReturn = uwVRMLConvOnExecuteStage1();
    if (auwReturn == RESULT_OK)
    {
      nvdSetStageControl(_VRMLCONV_STAGE2);
    }
    break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
  case _VRMLCONV_STAGE2:
    // time
    LARGE_INTEGER  freq, time_start, time_end;
    QueryPerformanceFrequency(&freq);
    QueryPerformanceCounter(&time_start);

    auwReturn = uwVRMLConvOnExecuteStage2();
    if (auwReturn == RESULT_OK)
    {
      std::cout << _VRMLCONVERT_SUCCESS << std::endl;
      nvdSetStageControl(_VRMLCONV_STAGE1);
    }
    else
    {
      nvdSetStageControl(ERROR_DETECTED);
    }

    // Process time
    QueryPerformanceCounter(&time_end);
    std::cout << "Process Time of Convert -> " << ((time_end.QuadPart - time_start.QuadPart) * 1000 / freq.QuadPart) << "[ms]" << std::endl;

    break;
    //---------------------------------------------------------------------------------

    //---------------------------------------------------------------------------------
  case ERROR_DETECTED:
    if (m_ubStageCtrlBackUp != ERROR_DETECTED)
    {
      std::cout << _VRMLCONVERT_FAILED << std::endl;
    }

    break;
    //---------------------------------------------------------------------------------

  default:
    if (m_ubStageCtrlBackUp != m_ubStageCtrl)
    {
      std::cout << _STAGE_ERROR << std::endl;
    }
    break;
  }

  m_ubStageCtrlBackUp = m_ubStageCtrl;
}

/***************************************************************************************************
*  Function        : nvdVRMLConvOnInitialize
*  Description     : When onInitialize, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnInitialize(void)
{
  std::cout << _ON_INITIALIZE << std::endl;
}

/***************************************************************************************************
*  Function        : nvdVRMLConvOnActivated
*  Description     : When onActivated, this function execute the initialization.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnActivated(void)
{
  std::cout << _ON_ACTIVATED << std::endl;

  nvdSetStageControl(_VRMLCONV_STAGE1);
  m_ubStageCtrlBackUp = _VRMLCONV_STAGE_INIT;

  nvdClrMultipleFlags();
}

/***************************************************************************************************
*  Function        : nvdVRMLConvOnDeactivated
*  Description     : When onDeactivated, this function execute the initialization.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnDeactivated(void)
{
  std::cout << _ON_DEACTIVATED << std::endl;
}

/***************************************************************************************************
*  Function        : nvdVRMLConvOnAborting
*  Description     : When onAborting, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnAborting(void)
{
  std::cout << _ON_ABORTING << std::endl;
  std::cout << _ERROR_RESET << std::endl;
}

/***************************************************************************************************
*  Function        : nvdVRMLConvOnReset
*  Description     : When onReset, this function outputs a message.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdVRMLConvOnReset(void)
{
  std::cout << _ON_RESET << std::endl;
}

/***************************************************************************************************
*  Function        : nvdSetStageControl
*  Description     : This function sets the member variable of m_ubStageCtrl.
*  Argument        : m_ubStageCtrl
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdSetStageControl(ubyte_type a_ubStageCtrl)
{
  if ((a_ubStageCtrl >= _VRMLCONV_STAGE_MIN) && (a_ubStageCtrl < _VRMLCONV_STAGE_MAX))
  {
    m_ubStageCtrl = a_ubStageCtrl;
  }
  else
  {
    m_ubStageCtrl = _VRMLCONV_STAGE1;
  }
}

/***************************************************************************************************
*  Function        : nvdClrMultipleFlags
*  Description     : This function clears multiple flags.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdClrMultipleFlags(void)
{
  m_calc_file_flag = FALSE;
  m_obj_file_flag = FALSE;
  m_img_file_flag = FALSE;
}

/***************************************************************************************************
*  Function        : sEditFilePath
*  Description     : This function edit the file path.
*  Argument        : file name and process mode
*  Return          : Returns a string of directory path or filename or extension.
**************************************************************************************************/
std::string rtc_3DMapVRMLConverter::sEditFilePath(std::string filename, uword_type mode)
{

  char_type strPathName[DEF_MAX_LENGTH];

  char_type strFullPath[DEF_MAX_LENGTH];
  char_type strDrive[DEF_MIN_LENGTH];
  char_type strDir[DEF_MAX_LENGTH];
  char_type strFile[DEF_MID_LENGTH];
  char_type strExt[DEF_MIN_LENGTH];

  memset(strFullPath, 0x00, sizeof(strFullPath));
  memset(strDrive, 0x00, sizeof(strDrive));
  memset(strDir, 0x00, sizeof(strDir));
  memset(strFile, 0x00, sizeof(strFile));
  memset(strExt, 0x00, sizeof(strExt));
  memset(strPathName, 0x00, sizeof(strPathName));

  //string -> char[]
  filename.copy(strFullPath, filename.size());
  _splitpath(strFullPath, strDrive, strDir, strFile, strExt);

  switch (mode)
  {
  case FILE_PATH_MODE_DIR:
    sprintf(strPathName, "%s%s", strDrive, strDir);
    break;

  case FILE_PATH_MODE_EXT:
    sprintf(strPathName, "%s", strExt);
    break;

  case FILE_PATH_MODE_FIL:
    sprintf(strPathName, "%s", strFile);
    break;

  case FILE_PATH_MODE_FNM:
  default:
    sprintf(strPathName, "%s%s", strFile, strExt);
    break;
  }

  return strPathName;
}

/***************************************************************************************************
*  Function        : uw3DMapVRMLConverterOnExecuteStage1
*  Description     : This function execute the main stage1 process of 3D Map VRMLC Converter RTC.
*  Argument        : nothing special
*  Return          : information of normal end or abnormal end.
**************************************************************************************************/
uword_type rtc_3DMapVRMLConverter::uwVRMLConvOnExecuteStage1(void)
{
  uword_type auwReturn = RESULT_NG;

  if (!m_calc_file_flag) nvdCheckofCalcFile();
  if (!m_obj_file_flag) nvdCheckofObjFile();
  if (!m_img_file_flag) nvdCheckofImgFile();

  if ((m_calc_file_flag == TRUE) && (m_obj_file_flag == TRUE) && (m_img_file_flag == TRUE))
  {
    nvdClrMultipleFlags();
    auwReturn = uwCheckFileRelevance();
  }

  return auwReturn;
}

/***************************************************************************************************
*  Function        : nvdCheckofCalcFile
*  Description     : This function checks the inputs calculation data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCheckofCalcFile(void)
{
  if (m_in_calcfilepathIn.isNew())
  {
    m_calc_file_flag = FALSE;
    m_in_calcfilepathIn.read();
    CHK_FILE_PARA chk_file_para = { _CHKP_CALC_TRG , _CHKP_CALC_MIN , _CHKP_CALC_MAX, { _CHKP_CALC_EXT} };

    if (bCheckInputFile(&m_in_calcfilepath, &chk_file_para))
    {
      nvdClrCalcResult();
      m_calc_file_flag = bSetCalcResult(&m_in_calcfilepath);
    }
  }
}

/***************************************************************************************************
*  Function        : nvdCheckofObjFile
*  Description     : This function checks the inputs object data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCheckofObjFile(void)
{
  if (m_in_objfilepathIn.isNew())
  {
    m_obj_file_flag = FALSE;
    m_in_objfilepathIn.read();
    CHK_FILE_PARA chk_file_para = { _CHKP_OBJ_TRG, _CHKP_OBJ_MIN, _CHKP_OBJ_MAX, { _CHKP_OBJ_EXT } };

    if (bCheckInputFile(&m_in_objfilepath, &chk_file_para))
    {
      nvdClrObjectData();
      m_obj_file_flag = bSetObjectData(&m_in_objfilepath);
    }
  }
}

/***************************************************************************************************
*  Function        : nvdCheckofImgFile
*  Description     : This function checks the inputs image data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCheckofImgFile(void)
{
  if (m_in_imgfilepathIn.isNew())
  {
    m_img_file_flag = FALSE;
    m_in_imgfilepathIn.read();
    CHK_FILE_PARA chk_file_para = { _CHKP_IMG_TRG, _CHKP_IMG_MIN, _CHKP_IMG_MAX, { _CHKP_IMG_EXT1
                                                                                 , _CHKP_IMG_EXT2
                                                                                 , _CHKP_IMG_EXT3
                                                                                 , _CHKP_IMG_EXT4
                                                                                 , _CHKP_IMG_EXT5
                                                                                 , _CHKP_IMG_EXT6
                                                                                 , _CHKP_IMG_EXT7
                                                                                 , _CHKP_IMG_EXT8 } };

    if (bCheckInputFile(&m_in_imgfilepath, &chk_file_para))
    {
      nvdClrImageFile();
      m_img_file_flag = bSetImageFile(&m_in_imgfilepath);
    }
  }
}

/***************************************************************************************************
*  Function        : bSetCalcResult
*  Description     : This function set the calculation data.
*  Argument        : sequential file name
*  Return          : Returns true if the calculation data is set, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bSetCalcResult(const RTC::TimedStringSeq *idp_filenameseq)
{
  m_calc_cnt = idp_filenameseq->data.length();

  for (uword_type i = 0; i < m_calc_cnt; i++)
  {
    std::string filename = idp_filenameseq->data[i];    

    if (!bReadCalcResultFile(filename , &m_calc_res[i]))
    {
      return FALSE;
    }
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : bReadTrajectFile
*  Description     : This function read the calculation file.
*  Argument        : file name and image buffer
*  Return          : Returns true if the file can be read, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bReadCalcResultFile(const std::string filename, CALC_RES *calc_res)
{
  bool_type abReturn = FALSE;

  std::cout << "Read the file(" << filename << ")" << std::endl;
    
  std::ifstream ifs(filename); //When instance is released, the file is closed automatically.
  if (!ifs.fail())
  {
    abReturn = bCalcResultSplitter(&ifs, calc_res);
  }
  else
  {
    /* File read error. */
    std::cout << _ERROR_FILE_READ_CALC << std::endl;
  }
  return abReturn;
}

/***************************************************************************************************
*  Function        : bCalcResultSplitter
*  Description     : This function splits the calculation file.
*  Argument        : input file stream and image buffer
*  Return          : Returns true if the file can be split, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCalcResultSplitter(std::ifstream *ifs, CALC_RES *calc_res)
{
  std::string str_line;
  std::vector<std::string> strFields;

  while (std::getline(*ifs, str_line))
  {
    /* Comment line goto next. */
    if (str_line.substr(0, 1) == _COMMENT_MARK) continue;

    strFields = split(str_line, _DELEMITER);
    if (strFields.size() == _FIELDS_CNT_CALC)
    {
      try
      {
        calc_res->af_x = stof(strFields[0]);
        calc_res->af_y = stof(strFields[1]);
        calc_res->var_z1 = stof(strFields[2]);
        calc_res->mean_z1 = stof(strFields[3]);
        calc_res->var_z2 = stof(strFields[4]);
        calc_res->mean_z2 = stof(strFields[5]);
        calc_res->dc = stof(strFields[6]);
      }
      catch (...)
      {
        /* Format error. */
        std::cout << _ERROR_FILE_FORMAT_CALC << std::endl;
        return FALSE;
      }
    }
    else
    {
      /* Format error. */
      std::cout << _ERROR_FILE_FORMAT_CALC << std::endl;
      return FALSE;
    }
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : bSetObjectData
*  Description     : This function set the object data.
*  Argument        : sequential file name
*  Return          : Returns true if the object data is set, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bSetObjectData(const RTC::TimedStringSeq *idp_filenameseq)
{
  m_obj_cnt = idp_filenameseq->data.length();

  for (uword_type i = 0; i < m_obj_cnt; i++)
  {
    std::string filename = idp_filenameseq->data[i];

    if (!bReadObjectDataFile(filename, &m_vec_vt[i], &m_vec_v[i]))
    {
      std::cout << "Error file : " << filename << std::endl;
      return FALSE;
    }

    if (!bCheckObjectData(&m_vec_vt[i]))
    {
      std::cout << "Error file : " << filename << std::endl;
      return FALSE;
    }
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : bReadObjectDataFile
*  Description     : This function read the object file.
*  Argument        : file name and image buffer
*  Return          : Returns true if the file can be read, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bReadObjectDataFile(const std::string filename, std::vector<LBL_VT> *obj_vt, std::vector<LBL_V> *obj_v)
{
  bool_type abReturn = FALSE;

  std::cout << "Read the file(" << filename << ")" << std::endl;

  std::ifstream ifs(filename); //When instance is released, the file is closed automatically.
  if (!ifs.fail())
  {
    abReturn = bObjectDataSplitter(&ifs, obj_vt, obj_v);
  }
  else
  {
    /* File read error. */
    std::cout << _ERROR_FILE_READ_OBJ << std::endl;
  }
  return abReturn;
}

/***************************************************************************************************
*  Function        : bObjectDataSplitter
*  Description     : This function splits the object file.
*  Argument        : input file stream and image buffer
*  Return          : Returns true if the file can be split, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bObjectDataSplitter(std::ifstream *ifs, std::vector<LBL_VT> *obj_vt, std::vector<LBL_V> *obj_v)
{
  std::string str_line;
  std::vector<std::string> strFields;

  LBL_VT vt_temp;
  LBL_V v_temp;

  while (std::getline(*ifs, str_line))
  {
    /* Comment line goto next. */
    if (str_line.substr(0, 1) == _COMMENT_MARK) continue;

    strFields = split(str_line, _DELEMITER);
    /* Empty line goto next. */
    if (strFields.empty()) continue;
    
    if (strFields[0] == _LABEL_VT)
    {
      if (strFields.size() == _FIELDS_CNT_OBJ_VT)
      {
        try
        {
          vt_temp.vt_x = stof(strFields[1]);
          vt_temp.vt_y = stof(strFields[2]);

          obj_vt->push_back(vt_temp);
        }
        catch (...)
        {
          /* Format error. */
          std::cout << _ERROR_FILE_FORMAT_OBJ << std::endl;
          return FALSE;
        }
      }
      else
      {
        /* Format error. */
        std::cout << _ERROR_FILE_FORMAT_OBJ << std::endl;
        return FALSE;
      }
    }

    if (strFields[0] == _LABEL_V)
    {
      if (strFields.size() == _FIELDS_CNT_OBJ_V)
      {
        try
        {
          v_temp.v_x = stof(strFields[1]); /* Dummy */
          v_temp.v_y = stof(strFields[2]); /* Dummy */
          v_temp.v_z = stof(strFields[3]) * -1;

          obj_v->push_back(v_temp);
        }
        catch (...)
        {
          std::cout << _ERROR_FILE_FORMAT_OBJ << std::endl;
          return FALSE;
        }
      }
      else
      {
        std::cout << _ERROR_FILE_FORMAT_OBJ << std::endl;
        return FALSE;
      }
    }
    
  }

  if (obj_vt->size() != obj_v->size())
  {
    std::cout << _ERROR_FILE_MISMAT_OBJ << std::endl;
    return FALSE;
  }
  
  return TRUE;
}

/***************************************************************************************************
*  Function        : bCheckObjectData
*  Description     : This function check the object data.
*  Argument        : image buffer
*  Return          : Returns true if the object data is correct, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCheckObjectData(const std::vector<LBL_VT> *vec_vt)
{   
  for (auto itr = vec_vt->begin(); itr != vec_vt->end(); ++itr)
  {
    if (itr->vt_x < _FLOAT_ZERO || _FLOAT_ONE < itr->vt_x || itr->vt_y < _FLOAT_ZERO || _FLOAT_ONE < itr->vt_y)
    {
      std::cout << _ERROR_AXIS_RANGE << std::endl;
      return FALSE;
    }    
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : bSetImageFile
*  Description     : This function set the image file path.
*  Argument        : sequential file name
*  Return          : Returns true if the image file path is set, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bSetImageFile(const RTC::TimedStringSeq *idp_filenameseq)
{
  m_img_cnt = idp_filenameseq->data.length();

  for (uword_type i = 0; i < m_img_cnt; i++)
  {
    m_img_file_path[i] = idp_filenameseq->data[i];

    if (!bCheckFileExist(m_img_file_path[i]))
    {
      std::cout << "Error file : " << m_img_file_path[i] << std::endl;
      return FALSE;
    }
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : split
*  Description     : This function splits the string.
*  Argument        : Line string and delimiter
*  Return          : Returns the string after split.
**************************************************************************************************/
std::vector<std::string> rtc_3DMapVRMLConverter::split(std::string& str_line, char_type delimiter)
{
  std::istringstream stream(str_line);
  std::string field;
  std::vector<std::string> result;

  while (getline(stream, field, delimiter))
  {
    result.push_back(field);
  }
  return result;
}

/***************************************************************************************************
*  Function        : bCheckCalcFile
*  Description     : This function checks the inputs filename.
*  Argument        : sequential file name
*  Return          : Returns true if the file exists and is correct, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCheckInputFile(RTC::TimedStringSeq *idp_filenameseq, CHK_FILE_PARA *chk_file_para)
{

  if (idp_filenameseq->data.length() < chk_file_para->min_seq_cnt || chk_file_para->max_seq_cnt < idp_filenameseq->data.length())
  {
    std::cout << _ERROR_FILE_CNT << chk_file_para->chk_target << std::endl;
    return FALSE;
  }

  for (uword_type i = 0; i < idp_filenameseq->data.length(); i++)
  {
    std::string filename = idp_filenameseq->data[i];
    if (!bCheckFileExist(filename))
    {
      std::cout << _ERROR_FILE_NAME << chk_file_para->chk_target  << std::endl;
      return FALSE;
    }

    if (!bCheckExtension(filename, &chk_file_para->list_ext))
    {
      std::cout << _ERROR_FILE_EXT << chk_file_para->chk_target << std::endl;
      return FALSE;
    }
  }

  return TRUE;

}

/***************************************************************************************************
*  Function        : bCheckFileExist
*  Description     : This function checks the existence of a file.
*  Argument        : file name
*  Return          : Returns true if the file exists, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCheckFileExist(std::string file_name)
{
  return (PathFileExists(file_name.c_str()));
}

/***************************************************************************************************
*  Function        : bCheckExtension
*  Description     : This function checks the extension.
*  Argument        : file path and target extension
*  Return          : Returns true if the file extension is correct, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCheckExtension(std::string filepath, std::list<std::string> *targetext)
{
  std::string strEXT = sEditFilePath(filepath, FILE_PATH_MODE_EXT);
  /* Upper -> Lower */
  transform(strEXT.begin(), strEXT.end(), strEXT.begin(), tolower);

  /* find extension */
  auto ite = std::find(targetext->begin(), targetext->end(), strEXT);
  if (ite != targetext->end())
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

/***************************************************************************************************
*  Function        : uwCheckFileRelevance
*  Description     : This function checks to file relevance.
*  Argument        : nothing special
*  Return          : Returns true if the number of file is correct, false otherwise.
**************************************************************************************************/
uword_type rtc_3DMapVRMLConverter::uwCheckFileRelevance(void)
{
  if (m_calc_cnt + 1 == m_obj_cnt && m_obj_cnt == m_img_cnt)
  {
    return RESULT_OK;
  }
  else
  {
    std::cout << _ERROR_FILE_RELEVANCE << std::endl;
    return RESULT_NG;
  }
}

/***************************************************************************************************
*  Function        : nvdClrCalcResult
*  Description     : This function clear calculation data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdClrCalcResult(void)
{
  m_calc_cnt = 0;
  for (uword_type i = 0; i < _CHKP_CALC_MAX; i++)
  {
    m_calc_res[i].af_x = _FLOAT_ZERO;
    m_calc_res[i].af_y = _FLOAT_ZERO;
    m_calc_res[i].var_z1 = _FLOAT_ZERO;
    m_calc_res[i].mean_z1 = _FLOAT_ZERO;
    m_calc_res[i].var_z2 = _FLOAT_ZERO;
    m_calc_res[i].mean_z2 = _FLOAT_ZERO;
    m_calc_res[i].dc = _FLOAT_ZERO;
  }
}

/***************************************************************************************************
*  Function        : nvdClrObjectData
*  Description     : This function clear object data.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdClrObjectData(void)
{
  m_obj_cnt = 0;
  for (uword_type i = 0; i < _CHKP_OBJ_MAX; i++)
  {    
    m_vec_vt[i].clear();
    m_vec_vt[i].shrink_to_fit();
    m_vec_v[i].clear();
    m_vec_v[i].shrink_to_fit();
  }
}

/***************************************************************************************************
*  Function        : nvdClrImageFile
*  Description     : This function clear image file path.
*  Argument        : nothing special
*  Return          : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdClrImageFile(void)
{
  m_img_cnt = 0;
  for (uword_type i = 0; i < _CHKP_IMG_MAX; i++)
  {
    m_img_file_path[i].clear();
    m_img_file_path[i].shrink_to_fit();
  }
}

/***************************************************************************************************
*  Function        : uwVRMLConvOnExecuteStage2
*  Description     : This function execute the main stage2 process of 3D Map VRMLC Converter RTC.
*  Argument        : nothing special
*  Return          : information of normal end or abnormal end.
**************************************************************************************************/
uword_type rtc_3DMapVRMLConverter::uwVRMLConvOnExecuteStage2(void)
{
  VRML_Z bf_z[_CHKP_OBJ_MAX] = {}; // Z before editing
  VRML_Z af_z[_CHKP_OBJ_MAX] = {}; // Z after editing
  DC_ACCUMULATE acc[_CHKP_OBJ_MAX] = {};
  float_type affine_x = _FLOAT_ZERO;
  float_type affine_y = _FLOAT_ZERO;
   
  std::stringstream strSaveFilename[_CHKP_OBJ_MAX];
  std::stringstream strSaveTexturename[_CHKP_OBJ_MAX];
  std::stringstream strSaveTexturepaht[_CHKP_OBJ_MAX];

  /* Timestamp */
  char date[32];
  time_t t = time(NULL);
  strftime(date, sizeof(date), "_%Y%m%d_%H%M%S_", localtime(&t));

  /* Make directory */
  if (!bCreateDirectory(_OUTPUT_DIRECTORY))
  {
    return RESULT_NG;
  }

  for (uword_type obj_cnt = 0; obj_cnt < m_obj_cnt; obj_cnt++)
  {
    /* Resolution conversion */
    nvdFittingObject(m_vec_vt[obj_cnt], m_vec_v[obj_cnt], &bf_z[obj_cnt]);

    /* Trimming */
    nvdTrimmingObject(&bf_z[obj_cnt], &af_z[obj_cnt]);
    
    /* Smoothing */
    nvdSmooth(&af_z[obj_cnt]);

    /* Calculation for mean and standard deviation */
    nvdCalcAccumulate(&af_z[obj_cnt], &acc[obj_cnt]);

    /* Calculation for moving objects */
    if (obj_cnt > 0)
    {
      /* Confirmation of enantiomorphism by distance */
      nvdCheckAndConvertEnantiomorphism(&af_z[obj_cnt - 1], &af_z[obj_cnt],
                                        acc[obj_cnt- 1].mean, acc[obj_cnt].mean, &m_calc_res[obj_cnt - 1]);

      /* Depth correction */
      nvdCorrection(obj_cnt, &af_z[obj_cnt], acc);
      /* Recalculation */
      nvdCalcAccumulate(&af_z[obj_cnt], &acc[obj_cnt]);
    
      /* Total affine */
      affine_x += (m_calc_res[obj_cnt - 1].af_x * COMPRESS_RATIO);
      affine_y += (m_calc_res[obj_cnt - 1].af_y * -1 * COMPRESS_RATIO);
    }    
    
    /* Edit and save texture */
    strSaveTexturename[obj_cnt] << "3DMap" << date << std::setw(2) << std::setfill('0') << obj_cnt + 1 << ".jpg";
    strSaveTexturepaht[obj_cnt] << _OUTPUT_DIRECTORY << "\\" << strSaveTexturename[obj_cnt].str();
    if (!bEditSaveTexture(m_img_file_path[obj_cnt], strSaveTexturepaht[obj_cnt].str()))
    {
      return RESULT_NG;
    }

    /* Save VRML file */
    strSaveFilename[obj_cnt] << _OUTPUT_DIRECTORY << "\\3DMap" << date << std::setw(2) << std::setfill('0') << obj_cnt + 1 << ".wrl";
    if (!bSaveVRMLFile(strSaveFilename[obj_cnt].str(), affine_x, affine_y, &af_z[obj_cnt], strSaveTexturename[obj_cnt].str()))
    {
      std::cout << _ERROR_FILE_SAVE_IMG << std::endl;
      return RESULT_NG;
    }
  }

  return RESULT_OK;
}

/***************************************************************************************************
*  Function        : bCreateDirectory
*  Description     : This function create the directory.
*  Argument        : target file path
*  Return          : Returns true if the directory was created, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bCreateDirectory(std::string filepath)
{
  (void)_mkdir(filepath.c_str());
  if (!bCheckFileExist(filepath))
  {
    std::cout << _ERROR_CREATE_DIRECTORY << std::endl;
    return FALSE;
  }
  return TRUE;
}

/***************************************************************************************************
*  Function        : nvdFittingObject
*  Description     : This function creates Z coordinate data.
*  Argument        : nothing special
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdFittingObject(const std::vector<LBL_VT> vec_vt, const std::vector<LBL_V> vec_v, VRML_Z *vrml)
{
  for (uword_type vec_cnt = 0; vec_cnt < vec_vt.size(); vec_cnt++)
  {
    uword_type x = static_cast<uword_type>(vec_vt[vec_cnt].vt_x * (XIMG_AF_CONV -1));
    uword_type y = static_cast<uword_type>(vec_vt[vec_cnt].vt_y * (YIMG_AF_CONV -1));

    vrml->z[x][y] = static_cast<float_type>(vec_v[vec_cnt].v_z * COMPRESS_RATIO);
  }
}

/***************************************************************************************************
*  Function        : nvdTrimmingObject
*  Description     : This function trimming Z coordinate data.
*  Argument        : Before Z coordinate and after Z coordinate.
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdTrimmingObject(const VRML_Z *bf_vrml, VRML_Z *af_vrml)
{
  uword_type min_x = 0, min_y = 0, max_x = 0, max_y = 0;  
  bool_type not_all_zero_min = FALSE;
  bool_type not_all_zero_max = FALSE;
  
  /* min x and man x*/
  not_all_zero_min = FALSE;
  not_all_zero_max = FALSE;
  for (uword_type cnt_u_x = 0, cnt_d_x = XIMG_AF_CONV - 1; cnt_u_x < XIMG_AF_CONV; cnt_u_x++, cnt_d_x--)
  {
    for (uword_type cnt_y = 0; cnt_y < YIMG_AF_CONV; cnt_y++)
    {
      if (!not_all_zero_min && bf_vrml->z[cnt_u_x][cnt_y] != _FLOAT_ZERO)
      {
        not_all_zero_min = TRUE;       
        min_x = cnt_u_x;
      }

      if (!not_all_zero_max && bf_vrml->z[cnt_d_x][cnt_y] != _FLOAT_ZERO)
      {
        not_all_zero_max = TRUE;
        max_x = cnt_d_x;
      }

      if (not_all_zero_min && not_all_zero_max) break;
    }

    if (not_all_zero_min && not_all_zero_max) break;
  }
    
  /* min y and max y */
  not_all_zero_min = FALSE;
  not_all_zero_max = FALSE;
  for (uword_type cnt_u_y = 0, cnt_d_y = YIMG_AF_CONV - 1; cnt_u_y < YIMG_AF_CONV; cnt_u_y++, cnt_d_y--)
  {
    for (uword_type cnt_x = 0; cnt_x < XIMG_AF_CONV; cnt_x++)
    {
      if (!not_all_zero_min && bf_vrml->z[cnt_x][cnt_u_y] != _FLOAT_ZERO)
      {
        not_all_zero_min = TRUE;
        min_y = cnt_u_y;
      }

      if (!not_all_zero_max && bf_vrml->z[cnt_x][cnt_d_y] != _FLOAT_ZERO)
      {
        not_all_zero_max = TRUE;
        max_y = cnt_d_y;
      }

      if (not_all_zero_min && not_all_zero_max) break;
    }

    if (not_all_zero_min && not_all_zero_max) break;
  }
  
  /* Trimming */
  uword_type from_y, from_x, to_x, to_y;
  for (from_y = min_y, to_y = 0; from_y <= max_y; from_y++, to_y++)
  {
    for (from_x = min_x, to_x = 0; from_x <= max_x; from_x++, to_x++)
    {
      af_vrml->z[to_x][to_y] = bf_vrml->z[from_x][from_y];
    }
  }

  /* length set */
  af_vrml->length_x = to_x;
  af_vrml->length_y = to_y;
}

/***************************************************************************************************
*  Function        : nvdSmooth
*  Description     : This function removes noise.
*  Argument        : nothing special
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdSmooth(VRML_Z *vrml)
{
  uword_type _height = vrml->length_y;
  uword_type _width = vrml->length_x;

  for (uword_type y = 1; y < (_height - 1); ++y)
  {
    for (uword_type x = 1; x < (_width - 1); ++x)
    {
      vrml->z[x][y] = static_cast<float_type>(
        vrml->z[x - 1][y - 1] * 0.0625 +
        vrml->z[x    ][y - 1] * 0.125  +
        vrml->z[x + 1][y - 1] * 0.0625 +
        vrml->z[x - 1][y    ] * 0.125  +
        vrml->z[x    ][y    ] * 0.25   +
        vrml->z[x + 1][y    ] * 0.125  +
        vrml->z[x - 1][y + 1] * 0.0625 +
        vrml->z[x    ][y + 1] * 0.125  +
        vrml->z[x + 1][y + 1] * 0.0625);
    }
  }

  for (uword_type x = 0; x < _width && 1 < _height; ++x)
  {
    vrml->z[x][_height - 1] = static_cast<float_type>(vrml->z[x][_height - 2] * 0.6 + vrml->z[x][_height - 1] * 0.4);
    vrml->z[x][0] = static_cast<float_type>(vrml->z[x][0] * 0.4 + vrml->z[x][1] * 0.6);
  }

  for (uword_type y = 0; y < _height && 1 < _width; ++y)
  {
    vrml->z[0][y] = static_cast<float_type>(vrml->z[0][y] * 0.4 + vrml->z[1][y] * 0.6);
    vrml->z[_width - 1][y] = static_cast<float_type>(vrml->z[_width - 1][y] * 0.4 + vrml->z[_width - 2][y] * 0.6);
  }
}

/***************************************************************************************************
*  Function        : nvdCalcAccumulate
*  Description     : This function calculates the mean and standard deviation of the Z coordinate.
*  Argument        : Z coordinate and accumulate data.
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCalcAccumulate(const VRML_Z *vrml, DC_ACCUMULATE *da)
{
  uword_type _height = vrml->length_y;
  uword_type _width = vrml->length_x;

  accumulator_set<float_type, stats<tag::variance>> acc;

  for (uword_type y = 0; y < _height; y++)
  {
    for (uword_type x = 0; x < _width; x++)
    {
      acc(vrml->z[x][y]);
    }
  }

  da->mean = static_cast<float_type>(extract::mean(acc));
  da->stdev = static_cast<float_type>(sqrt(extract::variance(acc)));
}

/***************************************************************************************************
*  Function        : nvdCheckAndConvertEnantiomorphism
*  Description     : This function check and convert the enantiomorphism.
*  Argument        : base and target Z coordinate and accumulate data and calculation result.
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCheckAndConvertEnantiomorphism(
                                                     const VRML_Z *base_z, VRML_Z *target_z, 
                                                     const float_type mean_base,
                                                     const float_type mean_target, CALC_RES *calc_res)
{
  uword_type base_x = 0, base_y = 0, target_x = 0, target_y = 0;
  uword_type count_org = 0, count_rev = 0;

  /* Move the index based on the calculation result */
  if (calc_res->af_x >= 0)
  {
    base_x = static_cast<uword_type>(calc_res->af_x * COMPRESS_RATIO);
  }
  else
  {
    target_x = static_cast<uword_type>(std::fabs(calc_res->af_x * COMPRESS_RATIO));
  }

  if (calc_res->af_y >= 0)
  {
    base_y = static_cast<uword_type>(calc_res->af_y * COMPRESS_RATIO);
  }
  else
  {
    target_y = static_cast<uword_type>(std::fabs(calc_res->af_y * COMPRESS_RATIO));
  }

  /* Judge whether it is enantiomeric pixel */
  for (uword_type y = 0; y < target_z->length_y; y++)
  {
    if (base_z->length_y <= y + base_y || target_z->length_y <= y + target_y)
    {
      continue;
    }

    for (uword_type x = 0; x < target_z->length_x; x++)
    {
      if (base_z->length_x <= x + base_x || target_z->length_x <= x + target_x)
      {
        break;
      }
      
      if (bIsEnantiomorphismPixel(base_z->z[x + base_x][y + base_y],
                                  target_z->z[x + target_x][y + target_y],
                                  mean_base, mean_target))
      {
        count_rev++;
      }
      else
      {
        count_org++;
      }
    }
  }

  if (count_org < count_rev)
  {
    /* Reverse the depth */
    nvdReverseDepth(target_z, mean_target);
  }
}

/***************************************************************************************************
*  Function        : nvdReverseDepth
*  Description     : This function reverses the Z coordinate.
*  Argument        : target Z coordinate and mean data.
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdReverseDepth(VRML_Z *target_z, float_type mean)
{
  for (uword_type y = 0; y < target_z->length_y; y++)
  {
    for (uword_type x = 0; x < target_z->length_x; x++)
    {
      target_z->z[x][y] = mean + (mean - target_z->z[x][y]);
    }
  }
}

/***************************************************************************************************
*  Function        : bIsEnantiomorphismPixel
*  Description     : This function checks the pixels enantiomorphism.
*  Argument        : Z coordinate and average data of base pixel and target pixel.
*  Argument        : In case of enantiomorphism, return true.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bIsEnantiomorphismPixel(const float_type pix_base_z,
                                                          const float_type pix_target_z , 
                                                          const float_type mean_base,
                                                          const float_type mean_target)
{
  bool_type positive_base = FALSE;
  bool_type positive_target = FALSE;

  if ((pix_base_z - mean_base) >= 0)
  {
    positive_base = TRUE;
  }
  
  if ((pix_target_z - mean_target) >= 0)
  {
    positive_target = TRUE;
  }

  if (positive_base != positive_target)
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

/***************************************************************************************************
*  Function        : nvdCorrection
*  Description     : This function corrects the Z coordinate with mean and standard deviation.
*  Argument        : index and Z coordinate and accumulate data.
*  Argument        : nothing special
**************************************************************************************************/
void rtc_3DMapVRMLConverter::nvdCorrection(const uword_type cnt, VRML_Z *af_z, DC_ACCUMULATE acc[_CHKP_OBJ_MAX])
{
  /* check division by zero */
  if (acc[cnt].stdev == _FLOAT_ZERO)
  {
    acc[cnt].dc = _FLOAT_ZERO;
  }
  else
  {
    /* Calculation for depth coefficient */
    acc[cnt].dc = acc[cnt - 1].stdev / acc[cnt].stdev;
  }

  for (uword_type y = 0; y < af_z->length_y; y++)
  {
    for (uword_type x = 0; x < af_z->length_x; x++)
    {
      af_z->z[x][y] = (af_z->z[x][y] - acc[cnt].mean) * acc[cnt].dc + acc[cnt - 1].mean;
    }
  }
}

/***************************************************************************************************
*  Function        : bEditSaveTexture
*  Description     : This function edits the image and saves it in a file.
*  Argument        : input file path and output file path
*  Return          : Returns true if the image file can be saved, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bEditSaveTexture(const std::string in_file_path, const std::string out_file_path)
{
  /* Read the file */
  cv::Mat im = cv::imread(in_file_path, cv::IMREAD_UNCHANGED);
  if (im.empty())
  {
    std::cout << _ERROR_FILE_FORMAT_IMG << std::endl;
    return FALSE;
  }

  /* Check color image */
  if (im.channels() != CHANNEL_NUM)
  {
    std::cout << _ERROR_FILE_NOT_COLOR << std::endl;
    return FALSE;
  }    

  /* Search blue back */
  uword_type min_x = 0, min_y = 0, max_x = 0, max_y = 0;
  if (!bSearchBlueback(&im, &min_x, &min_y, &max_x, &max_y))
  {
    std::cout << _ERROR_FILE_FORMAT_IMG << std::endl;
    return FALSE;
  }
  
  /* Trimming */
  cv::Mat tm(im, cv::Rect(min_x, min_y, max_x - min_x + 1, max_y - min_y + 1));

  /* Flip image */
  cv::Mat rm;
  flip(tm, rm, 0);
  
  /* Save image */
  if (!cv::imwrite(out_file_path, rm))
  {
    std::cout << _ERROR_FILE_SAVE_VRML << std::endl;
    return FALSE;
  }

  return TRUE;
}

/***************************************************************************************************
*  Function        : bSearchBlueback
*  Description     : This function searches for the blue color in the image.
*  Argument        : image matrix and searched point
*  Return          : Returns true if the searched, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bSearchBlueback(const cv::Mat *im, uword_type *min_x, uword_type *min_y, uword_type *max_x, uword_type *max_y)
{
  bool_type abReturn = FALSE;
  bool_type not_all_blue_min = FALSE;
  bool_type not_all_blue_max = FALSE;

  /* min x and max x */
  not_all_blue_min = FALSE;
  not_all_blue_max = FALSE;
  for (sword_type cnt_u_x = 0, cnt_d_x = im->cols - 1; cnt_u_x <= im->cols; cnt_u_x++, cnt_d_x--)
  {
    for (sword_type y = 0; y < im->rows; y++)
    {
      if (!not_all_blue_min && !(im->at<cv::Vec3b>(y, cnt_u_x)[RGB_R] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(y, cnt_u_x)[RGB_G] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(y, cnt_u_x)[RGB_B] == HIGH_LUMINANCE))
      {
        not_all_blue_min = TRUE;
        *min_x = static_cast<uword_type>(cnt_u_x);
      }

      if (!not_all_blue_max && !(im->at<cv::Vec3b>(y, cnt_d_x)[RGB_R] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(y, cnt_d_x)[RGB_G] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(y, cnt_d_x)[RGB_B] == HIGH_LUMINANCE))
      {
        not_all_blue_max = TRUE;
        *max_x = static_cast<uword_type>(cnt_d_x);
      }

      if (not_all_blue_min && not_all_blue_max) break;
    }

    if (not_all_blue_min && not_all_blue_max) break;
  }

  /* min y and max y */
  not_all_blue_min = FALSE;
  not_all_blue_max = FALSE;
  for (sword_type cnt_u_y = 0, cnt_d_y = im->rows - 1; cnt_u_y <= im->rows; cnt_u_y++, cnt_d_y--)
  {
    for (sword_type x = 0; x < im->cols; x++)
    {
      if (!not_all_blue_min && !(im->at<cv::Vec3b>(cnt_u_y, x)[RGB_R] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(cnt_u_y, x)[RGB_G] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(cnt_u_y, x)[RGB_B] == HIGH_LUMINANCE))
      {
        not_all_blue_min = TRUE;
        *min_y = static_cast<uword_type>(cnt_u_y);

      }

      if (!not_all_blue_max && !(im->at<cv::Vec3b>(cnt_d_y, x)[RGB_R] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(cnt_d_y, x)[RGB_G] == LOW_LUMINANCE &&
                                 im->at<cv::Vec3b>(cnt_d_y, x)[RGB_B] == HIGH_LUMINANCE))
      {
        not_all_blue_max = TRUE;
        *max_y = static_cast<uword_type>(cnt_d_y);

      }

      if (not_all_blue_min && not_all_blue_max) break;
    }

    if (not_all_blue_min && not_all_blue_max) break;
  }
  
  if (*min_x < *max_x && *min_y < *max_y)
  {
    abReturn = TRUE;
  }

  return abReturn;
}

/***************************************************************************************************
*  Function        : bSaveVRMLFile
*  Description     : This function saves the VRML in a file.
*  Argument        : file path and affine and vertex and texture
*  Return          : Returns true if the VRML file can be saved, false otherwise.
**************************************************************************************************/
bool_type rtc_3DMapVRMLConverter::bSaveVRMLFile(const std::string strFilename, const float_type affine_x, const float_type affine_y
                                              , const VRML_Z *vrml, const std::string strTexture)
{
  FILE* fp;

  if (strFilename == _EMPTY_STRING){
    return FALSE;
  }
    
  fp = fopen(strFilename.c_str(), "wb");
  if (fp == NULL){
    return FALSE;
  }

  fprintf(fp, "#VRML V2.0 utf8\n");
  fprintf(fp, "Transform {\n");
  fprintf(fp, "  translation %.6f %.6f 0\n", affine_x, affine_y);
  fprintf(fp, "  rotation 1 0 0 1.5708\n");
  fprintf(fp, "  children [\n");
  fprintf(fp, "    Shape {\n");
  fprintf(fp, "      geometry ElevationGrid {\n");
  fprintf(fp, "        xDimension %d\n", vrml->length_x);
  fprintf(fp, "        zDimension %d\n", vrml->length_y);
  fprintf(fp, "        xSpacing 1\n");
  fprintf(fp, "        zSpacing 1\n");
  fprintf(fp, "        height [\n");

  for (uword_type y = 0; y < vrml->length_y; y++)
  {
    fprintf(fp, "          ");
    for (uword_type x = 0; x < vrml->length_x; x++)
    {
      if (!(y == 0 && x == 0)) fprintf(fp, ",");
      fprintf(fp, "%.6f", vrml->z[x][y]);
    }
    fprintf(fp, "\n");
  }

  fprintf(fp, "        ]\n");
  fprintf(fp, "      }\n");
  fprintf(fp, "      appearance Appearance {\n");
  fprintf(fp, "        material Material {}\n");
  fprintf(fp, "        texture ImageTexture {\n");
  fprintf(fp, "        url \"%s\"\n", strTexture.c_str());
  fprintf(fp, "        }\n");
  fprintf(fp, "      }\n");
  fprintf(fp, "    }\n");
  fprintf(fp, "  ]\n");
  fprintf(fp, "}\n");

  fclose(fp);

  return TRUE;
}

extern "C"
{
 
  void rtc_3DMapVRMLConverterInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_3dmapvrmlconverter_spec);
    manager->registerFactory(profile,
                             RTC::Create<rtc_3DMapVRMLConverter>,
                             RTC::Delete<rtc_3DMapVRMLConverter>);
  }
  
};
